// #define _DEBUG

#include <math.h>
#include <limits.h>
#include "postgres.h"           // general Postgres declarations
#include "fmgr.h"               // Postgres function manager and function-call interface
#include "funcapi.h"            // for returning composite types and rows
#include "utils/typcache.h"     // for Type cache definitions
#include "access/tupmacs.h"     // Tuple macros used by both index tuples and heap tuples

#include "utils/array.h"        // Declarations for Postgres arrays.
// #include "contrib/intarray/_int.h"
// #include "executor/executor.h"  // for GetAttributeByName()

#define TIMESLOT_LENGTH_IN_SECONDS 0.5
#define MAX_SCORE 1.0f
#define MIN_DISTANCE_BETWEEN_OCCURRENCES 4

// identification... its me :)
#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

#ifdef _DEBUG
#define DBG(format, args...) ereport(INFO, (111111, errmsg(format, ## args)));
#else
#define DBG(format, args...) 
#endif

#define MIN(a, b) ((a)<(b)?(a):(b))
//#define INT32_MAX 2147483647

/*
 * The macro PG_ARGISNULL(n) allows a function to test whether each input is null. (Of course, 
 * doing this is only necessary in functions not declared "strict".) 
 * As with the PG_GETARG_xxx() macros, the input arguments are counted beginning at zero.
 * PG_GETARG_xxx_COPY() guarantees to return a copy of the specified argument that is safe for writing into.
 *
 * Never modify the contents of a pass-by-reference input value. If you do so you are likely to corrupt 
 * on-disk data, since the pointer you are given might point directly into a disk buffer. 
 * The sole exception to this rule is explained in Section 34.10 (User-Defined Aggregates). 
 * 
 * When allocating memory, use the PostgreSQL functions palloc and pfree instead of the corresponding 
 * C library functions malloc and free. The memory allocated by palloc will be freed automatically 
 * at the end of each transaction, preventing memory leaks. 
 * Always zero the bytes of your structures using memset. Without this, it's difficult to support 
 * hash indexes or hash joins, as you must pick out only the significant bits of your data structure 
 * to compute a hash. Even if you initialize all fields of your structure, there might be alignment padding 
 * (holes in the structure) that contain garbage values.
 *
 */



#define MAX_DIST 5
#define MIN_DIST 0


// c_srch_pos_near_weights() {{{
PG_FUNCTION_INFO_V1(c_srch_pos_near_weights);
/****************************************************************************************************
 * For a query consisting of several words, it takes positions and probabilities of the words
 * in one document and computes the number of occurrences of the whole query inside that document.
 * It also computes the document's score for ranking the search results.
 *
 * @param positions1 int4[] positions of the first word in query
 * @param weights1 float4[] probabilities of the first word in query
 * @param positions2 int4[] ... second word ...
 * @param weights2 float4[] ... second word ...
 * ...
 * @param positionsN int4[] ... Nth word ...
 * @param weightsN float4[] ... Nth word ...  
 */
Datum 
c_srch_pos_near_weights(PG_FUNCTION_ARGS) 
{
	int         j; // loop counter

	int32       vectors_count;
	int32      *len;
	int32*     *v;
	float4*    *w;
	int32      *i;
	int32      *hit;
	int32      *first_hit_i;

	int32       hit_start_boundary;
	int32       hit_end_boundary;
	int32       hit_best_start;
	int32       hit_best_end;
	float4      hit_best_score;

	int32       a;
	int32       b;

	int32       hits_count;
	float4      score;
	bool 		done;

	vectors_count = (int)(PG_NARGS()/2);
	DBG("vectors_count: %d", vectors_count);

	len         = palloc(sizeof(int32*)  * vectors_count);
	v           = palloc(sizeof(int32*)  * vectors_count);
	w           = palloc(sizeof(float4*) * vectors_count);
	i           = palloc(sizeof(int32*)  * vectors_count);
	hit         = palloc(sizeof(int32*)  * vectors_count);
	first_hit_i = palloc(sizeof(int32*)  * vectors_count);

	int arg_counter = 0;
	for (j=0; j<vectors_count; j++) {
		ArrayType* vector_pos = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_DATUM(arg_counter++)));
		ArrayType* vector_weight = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_DATUM(arg_counter++)));

		len[j] = ArrayGetNItems(ARR_NDIM(vector_pos), ARR_DIMS(vector_pos));
		DBG("len[%d] = %d", j, len[j]);
		v[j]   = (int32*) ARR_DATA_PTR(vector_pos);         // array data pointers
		w[j]   = (float4*) ARR_DATA_PTR(vector_weight);         // array data pointers
		if ( len[j] > ArrayGetNItems(ARR_NDIM(vector_weight), ARR_DIMS(vector_weight)) ) {
			ereport(ERROR, (errcode(ERRCODE_CARDINALITY_VIOLATION),
						errmsg("weight array %d must be of the same size as position array %d", j, j)));
		}
	}

	for (j=0; j<vectors_count; j++) { i[j] = 0; }

	for (j=0; j<vectors_count; j++) { hit[j] = 0; }

	for (j=0; j<vectors_count; j++) { first_hit_i[j] = -1; }

	hit_start_boundary = -1;
	hit_end_boundary   = -1;
	hit_best_start     = -1;
	hit_best_end       = -1;
	hit_best_score     =  0;
	hits_count         =  0;
	score              =  0;
	a                  =  0;
	b                  =  1;
	done               =  false;

	while(1) {
		DBG("    a=%d b=%d", a,b);
		int dist = v[b][i[b]] - v[a][i[a]];
		DBG("    distance(%d, %d) = %d", v[a][i[a]], v[b][i[b]], dist);
        if (dist <= MAX_DIST && dist >= MIN_DIST) {
			if (first_hit_i[b] == -1) {
				first_hit_i[b] = i[b];
				DBG("    first_hit_i[%d] = %d", b, v[b][i[b]]);
			}
			if (b == vectors_count-1) {
				float hit_cur_score = MAX_SCORE;
				for (j=0; j<vectors_count; j++) {
					if (hit_cur_score > w[j][i[j]]) {
						hit_cur_score = w[j][i[j]];
					}
				}
				DBG("      HIT SCORE: %f", hit_cur_score);
				//DBG("      HIT: i[%d,%d,%d] -> [%d %d %d]", i[0],i[1],i[2], v[0][i[0]],v[1][i[1]],v[2][i[2]]);
				// if the current hit does not overlap with the last hit
				if (hit_end_boundary < v[0][i[0]]) {
					if (hit_best_start != -1) {
						DBG("HIT %d..%d score:%f", v[0][hit_best_start], v[b][hit_best_end], hit_best_score);
						score += hit_best_score; // first time it will add zero, which does not matter
						hits_count++;
					}
					DBG("      NEW HIT");
					hit_best_score = hit_cur_score;
					hit_best_start = i[0];
					hit_best_end = i[b];
					hit_start_boundary = v[0][i[0]];
					hit_end_boundary = v[b][i[b]];
				} else {
					DBG("      OVERLAPPING HIT");
					if (hit_end_boundary < v[b][i[b]]) {
						hit_end_boundary = v[b][i[b]];
					}
					if (hit_best_score < hit_cur_score) {
						hit_best_score = hit_cur_score;
						hit_best_start = i[0];
						hit_best_end = i[b];
					}
				}
				i[b]++;
				while (i[b] >= len[b]) {
					if (a==0) { break; }
					i[b] = first_hit_i[b];
					i[a]++;
					a--; b--;
				}
			} else {
				a++; b++;
			}
		} else {
			DBG("    v[%d][%d] <> v[%d][%d] ... %d <> %d", a,i[a], b,i[b], v[a][i[a]], v[b][i[b]]);
			if (v[b][i[b]] < v[a][i[a]]) {
				if (first_hit_i[b] != -1) {
					first_hit_i[b] = -1;
					DBG("    first_hit_i[%d] = -1", b);
				}
				i[b]++;
				DBG("    i[%d]++ -> %d", b, v[b][i[b]]);
				if (i[b] >= len[b]) { break; } // now we are sure that there are no more hits
			} else {
				if (first_hit_i[b] != -1) {
					i[b] = first_hit_i[b];
				} else {
					i[b]--;
				}
				i[a]++;
				if (i[a] >= len[a]) { break; } // we reached the end of the first vector
				DBG("    i[%d] -> %d", b, v[b][i[b]]);
				DBG("    i[%d]++ -> %d", a, v[a][i[a]]);
				if (a!=0) {
					a--; b--;
				}
			}
		}
	}
	if (hit_best_start != -1) {
		DBG("HIT %d..%d score:%f", v[0][hit_best_start], v[b][hit_best_end], hit_best_score);

		score += hit_best_score;
		hits_count++;
		done = true;
		DBG("score += %f (%f)", hit_best_score, score);
	}
	DBG("hits_count: %d, score: %f", hits_count, score);
//	PG_RETURN_FLOAT4(score);



	TupleDesc tupdesc;
	HeapTuple resultTuple;
	Datum values[2];
	bool  nulls[] = {0, 0};

	if( get_call_result_type( fcinfo, NULL, &tupdesc ) != TYPEFUNC_COMPOSITE )
		ereport( ERROR,
				( errcode( ERRCODE_FEATURE_NOT_SUPPORTED ),
				  errmsg( "function returning record called in context that cannot accept type record" )));

	if(tupdesc->natts != 2)
		ereport( ERROR,
				( errcode( ERRCODE_FEATURE_NOT_SUPPORTED ),
				  errmsg( "function returns record with 2 members ")));

	values[0] = UInt32GetDatum(hits_count);
	values[1] = Float4GetDatum(score);

	BlessTupleDesc( tupdesc );

	resultTuple = heap_form_tuple(tupdesc, values, nulls);
	PG_RETURN_DATUM( HeapTupleGetDatum(resultTuple) );
}
// }}}

// c_srch_pos_near_weights_oneword() {{{
PG_FUNCTION_INFO_V1(c_srch_pos_near_weights_oneword);
/****************************************************************************************************
 * Similar to c_srch_pos_near_weights but this one is for searching single word queries.
 *
 * @param positions1 int4[] positions of the word in query
 * @param weights1 float4[] probabilities of the word in query
 */

Datum c_srch_pos_near_weights_oneword(PG_FUNCTION_ARGS) 
{
	int         j; // loop counter

	int32       vectors_count;
	int32       len;
	int32*      v;
	float4*     w;
	int32       i;
	int32       hit;
	int32       first_hit_i;

	int32       hits_count;
	float4      score;

	ArrayType* vector_pos = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_DATUM(0)));
	ArrayType* vector_weight = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_DATUM(1)));

	len = ArrayGetNItems(ARR_NDIM(vector_pos), ARR_DIMS(vector_pos));
	DBG("len = %d", len);
	v   = (int32*) ARR_DATA_PTR(vector_pos);         // array data pointers
	w   = (float4*) ARR_DATA_PTR(vector_weight);         // array data pointers
	if ( len > ArrayGetNItems(ARR_NDIM(vector_weight), ARR_DIMS(vector_weight)) ) {
		ereport(ERROR, (errcode(ERRCODE_CARDINALITY_VIOLATION),
					errmsg("weight array must be of the same size as position array")));
	}

	hits_count         =  0;
	score              =  0;
	int32 hit_start    = v[0];
	float4 hit_score   = w[0];

	while(i < len) {
		DBG("i:%d hit_start:%d", i, v[i]);
//        if (i==len-1 || v[i+1] > v[i]+1 || w[i+1] != w[i]) {
        if (i==len-1 || v[i+1]-hit_start > MIN_DISTANCE_BETWEEN_OCCURRENCES) {
			score += hit_score < w[i] ? w[i] : hit_score;
			hits_count ++;
			hit_score = w[i+1];
			hit_start = v[i+1];
		} else {
			if (hit_score < w[i]) {
				hit_score = w[i];
			}
		}
		i++;
	}


	TupleDesc tupdesc;
	HeapTuple resultTuple;
	Datum values[2];
	bool  nulls[] = {0, 0};

	if( get_call_result_type( fcinfo, NULL, &tupdesc ) != TYPEFUNC_COMPOSITE )
		ereport( ERROR,
				( errcode( ERRCODE_FEATURE_NOT_SUPPORTED ),
				  errmsg( "function returning record called in context that cannot accept type record" )));

	if(tupdesc->natts != 2)
		ereport( ERROR,
				( errcode( ERRCODE_FEATURE_NOT_SUPPORTED ),
				  errmsg( "function returns record with 2 members ")));

	values[0] = UInt32GetDatum(hits_count);
	values[1] = Float4GetDatum(score);

	BlessTupleDesc( tupdesc );

	resultTuple = heap_form_tuple(tupdesc, values, nulls);
	PG_RETURN_DATUM( HeapTupleGetDatum(resultTuple) );
}
// }}}

typedef struct search_state
{
	TupleDesc	tupd;

	int32       vectors_count;
	int32      *len;
	int32*     *v;
	float4*    *w;
	int32      *i;
	int32      *hit;
	int32      *first_hit_i;

	int32       hit_start_boundary;
	int32       hit_end_boundary;
	int32       hit_best_start;
	int32       hit_best_end;
	float4      hit_best_score;

	int32       a;
	int32       b;

	float4      score;
	bool 		done;
} search_state;

// c_srch_pos_near_weights_occurrences() {{{
PG_FUNCTION_INFO_V1(c_srch_pos_near_weights_occurrences);
/****************************************************************************************************
 * For a query consisting of several words, it takes positions and probabilities of the words
 * in one document and produces a list of occurrences of the whole query inside that document
 * containing start and end times and probabilities of the occurrences.
 *
 * @param positions1 int4[] positions of the first word in query
 * @param weights1 float4[] probabilities of the first word in query
 * @param positions2 int4[] ... second word ...
 * @param weights2 float4[] ... second word ...
 * ...
 * @param positionsN int4[] ... Nth word ...
 * @param weightsN float4[] ... Nth word ...  
 */
Datum 
c_srch_pos_near_weights_occurrences(PG_FUNCTION_ARGS) 
{
	FuncCallContext *funcctx;
	int call_cntr;
	int max_calls;
	search_state *s = NULL;
	Datum        result;
	int j; // loop counter

	if (SRF_IS_FIRSTCALL()) {
		MemoryContext oldcontext;

		funcctx = SRF_FIRSTCALL_INIT();
		// switch to memory context appropriate for multiple function calls 
		oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
		funcctx->max_calls = INT32_MAX;
		// allocate the state object
		s = palloc(sizeof(search_state));

		s->vectors_count = (int)(PG_NARGS()/2);
		DBG("vectors_count: %d", s->vectors_count);

		s->len         = palloc(sizeof(int32*)  * s->vectors_count);
		s->v           = palloc(sizeof(int32*)  * s->vectors_count);
		s->w           = palloc(sizeof(float4*) * s->vectors_count);
		s->i           = palloc(sizeof(int32*)  * s->vectors_count);
		s->hit         = palloc(sizeof(int32*)  * s->vectors_count);
		s->first_hit_i = palloc(sizeof(int32*)  * s->vectors_count);

		int arg_counter = 0;
		for (j=0; j<s->vectors_count; j++) {
			ArrayType* vector_pos = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_DATUM(arg_counter++)));
			ArrayType* vector_weight = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_DATUM(arg_counter++)));

			s->len[j] = ArrayGetNItems(ARR_NDIM(vector_pos), ARR_DIMS(vector_pos));
			DBG("s->len[%d] = %d", j, s->len[j]);
			s->v[j]   = (int32*) ARR_DATA_PTR(vector_pos);         // array data pointers
			s->w[j]   = (float4*) ARR_DATA_PTR(vector_weight);         // array data pointers
			if ( s->len[j] > ArrayGetNItems(ARR_NDIM(vector_weight), ARR_DIMS(vector_weight)) ) {
				ereport(ERROR, (errcode(ERRCODE_CARDINALITY_VIOLATION),
							errmsg("weight array %d must be of the same size as position array %d", j, j)));
			}
			if (funcctx->max_calls > s->len[j]) {
				funcctx->max_calls = s->len[j];
			}
		}

		for (j=0; j<s->vectors_count; j++) { s->i[j] = 0; }

		for (j=0; j<s->vectors_count; j++) { s->hit[j] = 0; }

		for (j=0; j<s->vectors_count; j++) { s->first_hit_i[j] = -1; }

		s->hit_start_boundary = -1;
		s->hit_end_boundary   = -1;
		s->hit_best_start     = -1;
		s->hit_best_end       = -1;
		s->hit_best_score     =  0;
		s->score              =  0;
		s->a                  =  0;
		s->b                  =  1;
		s->done               =  false;

		if (get_call_result_type(fcinfo, NULL, &(s->tupd)) != TYPEFUNC_COMPOSITE) {
			ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED), errmsg
						("function returning record called in context "
						 "that cannot accept type record")));
		}

		funcctx->user_fctx = s;

		MemoryContextSwitchTo(oldcontext);
	}

	funcctx = SRF_PERCALL_SETUP();

	s = funcctx->user_fctx;

	if (s->done) {
		DBG("SRF_RETURN_DONE()");
		SRF_RETURN_DONE(funcctx);
	}

	int32       *len = s->len;
	int32*      *v   = s->v;
	float4*     *w   = s->w;
	int32       *i   = s->i;
	int32        a   = s->a;
	int32        b   = s->b;
	int32       *first_hit_i = s->first_hit_i;

	while(1) {
		DBG("    a=%d b=%d", a,b);
		int dist = v[b][i[b]] - v[a][i[a]];
		DBG("    distance(%d, %d) = %d", v[a][i[a]], v[b][i[b]], dist);
        if (dist <= MAX_DIST && dist >= MIN_DIST) {
			if (first_hit_i[b] == -1) {
				first_hit_i[b] = i[b];
				DBG("    first_hit_i[%d] = %d", b, v[b][i[b]]);
			}
			if (b == s->vectors_count-1) {
				float hit_cur_score = MAX_SCORE;
				for (j=0; j<s->vectors_count; j++) {
					if (hit_cur_score > w[j][i[j]]) {
						hit_cur_score = w[j][i[j]];
					}
				}
				DBG("      HIT SCORE: %f", hit_cur_score);
				//DBG("      HIT: i[%d,%d,%d] -> [%d %d %d]", i[0],i[1],i[2], v[0][i[0]],v[1][i[1]],v[2][i[2]]);
				// if the current hit does not overlap with the last hit
				if (s->hit_end_boundary < v[0][i[0]]) {
					if (s->hit_best_start != -1) {
						__c_srch_pos_near_weights_occurrences_create_result_obj(&s->tupd, v[0][s->hit_best_start], v[b][s->hit_best_end], s->hit_best_score, &result);
						DBG("HIT %d..%d score:%f", v[0][s->hit_best_start], v[b][s->hit_best_end], s->hit_best_score);
						s->score += s->hit_best_score; // first time it will add zero, which does not matter

						// store variables in the state structure before calling SRF_RETURN_NEXT() {{{
						s->hit_best_score = hit_cur_score;
						s->hit_best_start = i[0];
						s->hit_best_end = i[b];
						s->hit_start_boundary = v[0][i[0]];
						s->hit_end_boundary = v[b][i[b]];

						i[b]++;
						while (i[b] >= len[b]) {
							if (a==0) { break; }
							i[b] = first_hit_i[b];
							i[a]++;
							a--; b--;
						}

						s->a = a;
						s->b = b;
						// }}}

						DBG("SRF_RETURN_NEXT()");
						SRF_RETURN_NEXT(funcctx, result); // this call returns from the function like 'return result;'
					}
					DBG("      NEW HIT");
					s->hit_best_score = hit_cur_score;
					s->hit_best_start = i[0];
					s->hit_best_end = i[b];
					s->hit_start_boundary = v[0][i[0]];
					s->hit_end_boundary = v[b][i[b]];
				} else {
					DBG("      OVERLAPPING HIT");
					if (s->hit_end_boundary < v[b][i[b]]) {
						s->hit_end_boundary = v[b][i[b]];
					}
					if (s->hit_best_score < hit_cur_score) {
						s->hit_best_score = hit_cur_score;
						s->hit_best_start = i[0];
						s->hit_best_end = i[b];
					}
				}
				i[b]++;
				while (i[b] >= len[b]) {
					if (a==0) { break; }
					i[b] = first_hit_i[b];
					i[a]++;
					a--; b--;
				}
			} else {
				a++; b++;
			}
		} else {
			DBG("    v[%d][%d] <> v[%d][%d] ... %d <> %d", a,i[a], b,i[b], v[a][i[a]], v[b][i[b]]);
			if (v[b][i[b]] < v[a][i[a]]) {
				if (first_hit_i[b] != -1) {
					first_hit_i[b] = -1;
					DBG("    first_hit_i[%d] = -1", b);
				}
				i[b]++;
				DBG("    i[%d]++ -> %d", b, v[b][i[b]]);
				if (i[b] >= len[b]) { DBG("break"); break; } // now we are sure that there are no more hits
			} else {
				if (first_hit_i[b] != -1) {
					i[b] = first_hit_i[b];
				} else {
					i[b]--;
				}
				i[a]++;
				if (i[a] >= len[a]) { DBG("break"); break; } // we reached the end of the first vector
				DBG("    i[%d] -> %d", b, v[b][i[b]]);
				DBG("    i[%d]++ -> %d", a, v[a][i[a]]);
				if (a!=0) {
					a--; b--;
				}
			}
		}
	}
	if (s->hit_best_start != -1) {
		DBG("HIT %d..%d score:%f", v[0][s->hit_best_start], v[b][s->hit_best_end], s->hit_best_score);

		s->score += s->hit_best_score;
		s->done = true;
		DBG("score += %f (%f)", s->hit_best_score, s->score);
		__c_srch_pos_near_weights_occurrences_create_result_obj(&s->tupd, v[0][s->hit_best_start], v[b][s->hit_best_end], s->hit_best_score, &result);
		//__c_srch_pos_near_weights_occurrences_create_result_obj(&s->tupd, v1[s->i_prev_hit], v2[s->j_prev_hit], s->score_current_best, &result);
		DBG("SRF_RETURN_NEXT()");
		SRF_RETURN_NEXT(funcctx, result); // this call returns from the function like 'return result;'
	}

	DBG("SRF_RETURN_DONE()");
	SRF_RETURN_DONE(funcctx);
}
// }}}

typedef struct search_state_oneword
{
	TupleDesc	tupd;

	int32      len;
	int32*     v;
	float4*    w;
	int32      i;

	bool 		done;
} search_state_oneword;

// c_srch_pos_near_weights_occurrences_oneword() {{{
PG_FUNCTION_INFO_V1(c_srch_pos_near_weights_occurrences_oneword);
/****************************************************************************************************
 * Similar to c_srch_pos_near_weights_occurrences but this one is for searching single word queries.
 *
 * @param positions1 int4[] positions of the word in query
 * @param weights1 float4[] probabilities of the word in query
 */
Datum 
c_srch_pos_near_weights_occurrences_oneword(PG_FUNCTION_ARGS) 
{
	FuncCallContext *funcctx;
	int call_cntr;
	int max_calls;
	search_state_oneword *s = NULL;
	Datum        result;
	int j; // loop counter

	if (SRF_IS_FIRSTCALL()) {
		MemoryContext oldcontext;

		funcctx = SRF_FIRSTCALL_INIT();
		// switch to memory context appropriate for multiple function calls 
		oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
		funcctx->max_calls = INT32_MAX;
		// allocate the state object
		s = palloc(sizeof(search_state_oneword));

		ArrayType* vector_pos = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_DATUM(0)));
		ArrayType* vector_weight = (ArrayType *) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_DATUM(1)));

		s->len = ArrayGetNItems(ARR_NDIM(vector_pos), ARR_DIMS(vector_pos));
		DBG("s->len = %d", s->len);
		s->v   = (int32*) ARR_DATA_PTR(vector_pos);         // array data pointers
		s->w   = (float4*) ARR_DATA_PTR(vector_weight);         // array data pointers
		if ( s->len > ArrayGetNItems(ARR_NDIM(vector_weight), ARR_DIMS(vector_weight)) ) {
			ereport(ERROR, (errcode(ERRCODE_CARDINALITY_VIOLATION),
						errmsg("weight array must be of the same size as position array")));
		}
		funcctx->max_calls = s->len;

		s->i     = 0;
		s->done  = false;

		if (get_call_result_type(fcinfo, NULL, &(s->tupd)) != TYPEFUNC_COMPOSITE) {
			ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED), errmsg
						("function returning record called in context "
						 "that cannot accept type record")));
		}

		funcctx->user_fctx = s;

		MemoryContextSwitchTo(oldcontext);
	}

	funcctx = SRF_PERCALL_SETUP();

	s = funcctx->user_fctx;

	if (s->done) {
		DBG("SRF_RETURN_DONE()");
		SRF_RETURN_DONE(funcctx);
	}

	int32       len = s->len;
	int32*      v   = s->v;
	float4*     w   = s->w;
	int32       i   = s->i;

	int32       hit_start = v[i];
	float4      hit_score = w[i];

	while(i < len) {
		DBG("i:%d hit_start:%d", i, v[i]);
//        if (i==len-1 || v[i+1] > v[i]+1 || w[i+1] != w[i]) {
        if (i==len-1 || v[i+1]-hit_start > MIN_DISTANCE_BETWEEN_OCCURRENCES) {
			__c_srch_pos_near_weights_occurrences_create_result_obj(&s->tupd, hit_start, v[i], hit_score < w[i] ? w[i] : hit_score, &result);
			//DBG("HIT %d..%d score:%f", v[0][s->hit_best_start], v[b][s->hit_best_end], s->hit_best_score);
			s->i = i+1;

			DBG("SRF_RETURN_NEXT()");
			SRF_RETURN_NEXT(funcctx, result); // this call returns from the function like 'return result;'
		} else {
			if (hit_score < w[i]) {
				hit_score = w[i];
			}
			i++;
		}
	}

	DBG("SRF_RETURN_DONE()");
	SRF_RETURN_DONE(funcctx);
}
// }}}


